using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using SimpleJSON;
using GPUTools.Physics.Scripts.Behaviours;
using System.Linq;

public class FakeCollider 
    {
        // Atom atom;
        GameObject parent;
        CapsuleCollider ccollider;
        CapsuleLineSphereCollider cgpu;
        public DAZBone anchor;
        public DAZBone anchor2; // allow for binding to average of 2 anchors
        GameObject preview;
        Vector3 center;
        Vector3 rotation;
        JSONStorableFloat radius;
        JSONStorableFloat height;
        JSONStorableFloat x;
        JSONStorableFloat y;
        JSONStorableFloat z;
        JSONStorableFloat rx;
        JSONStorableFloat ry;
        JSONStorableFloat rz;
        UIDynamicSlider s_radius;
        UIDynamicSlider s_height;
        UIDynamicSlider s_x;
        UIDynamicSlider s_y;
        UIDynamicSlider s_z;
        UIDynamicSlider s_rx;
        UIDynamicSlider s_ry;
        UIDynamicSlider s_rz;

        FakeCollider interp_n1;  // exclusively for interpolated colliders
        FakeCollider interp_n2; //  ^^
        bool isInterpA;

        bool isInterpolated;

        List<string> ignoredRigidBodies;
        Vector3 originalCenterOfMass;
        Vector3 originalInertiaTensor;
        Quaternion originalInertiaTensorRotation;

        bool showpreview = false;
        bool gui = false;
        public bool seethrough=true;
        // bool issphere;
        public string name;
        
        public static float seethroughalpha=0.05f;
        public static float occludedalpha=0.3f;
        
        public FakeCollider(UYCol scr, string n, string anchorid, List<string> ignoreRBs, FakeCollider i1 = null, FakeCollider i2 = null, bool interpA=true)
        {
            name = n;
            center = new Vector3(0,0,0);
            rotation = new Vector3(0,0,0);
            interp_n1=i1;
            interp_n2=i2;
            isInterpolated = false;
            ignoredRigidBodies = ignoreRBs;
            if ((i1!=null) && (i2!=null)) {
              List<string> combined = i1.ignoredRigidBodies;
              combined.AddRange(i2.ignoredRigidBodies);
              ignoredRigidBodies = combined; 
              isInterpolated=true;
              isInterpA = interpA;
            }



            Atom containingAtom = scr.GetContainingAtom();
            string[] anchorids = anchorid.Split('#');
            JSONStorable _anchor = containingAtom.GetStorableByID(anchorids[0]);
            anchor = _anchor as DAZBone;
            anchor2 = null;
            if (anchorids.Length>1)
            {
                _anchor = containingAtom.GetStorableByID(anchorids[1]);
                anchor2 = _anchor as DAZBone;
            }


            parent = new GameObject();
            parent.name = name;
            ccollider = parent.AddComponent<CapsuleCollider>();
            ccollider.radius = 0.1f;
            ccollider.height = 0.2f;
            cgpu = parent.AddComponent<CapsuleLineSphereCollider>();
            cgpu.capsuleCollider = ccollider;
            cgpu.UpdateData();



            initSliders(scr);
            ccollider.isTrigger = true;
            ccollider.gameObject.transform.SetParent(parent.transform);
            parent.transform.SetParent(anchor.transform);
            
            originalCenterOfMass = ccollider.attachedRigidbody.centerOfMass;
            originalInertiaTensor = ccollider.attachedRigidbody.inertiaTensor;
            originalInertiaTensorRotation = ccollider.attachedRigidbody.inertiaTensorRotation;
            var allColliders = containingAtom.GetComponentsInChildren<CapsuleCollider>();

            foreach (CapsuleCollider c in allColliders) {
              if (ignoredRigidBodies.Any((s)=>c.attachedRigidbody.name.Contains(s))) {
                UnityEngine.Physics.IgnoreCollision(ccollider, c);
              }
            }


        }

        public void SetTrigger(List<string> rigidbodyFilter, bool v) {
          if (rigidbodyFilter.Any((filter)=> ccollider.attachedRigidbody.name.Contains(filter))) {          
            ccollider.isTrigger = v;
            ccollider.attachedRigidbody.centerOfMass = originalCenterOfMass;
            ccollider.attachedRigidbody.inertiaTensor = originalInertiaTensor;
            ccollider.attachedRigidbody.inertiaTensorRotation = originalInertiaTensorRotation;
          }
        } 

        public void initSliders(UYCol scr)
        {
            radius = new JSONStorableFloat("radius",ccollider.radius, (val) => 
                { 
                    if (val > ccollider.height/2) {
                        val = ccollider.height/2;
                        radius.valNoCallback = val;
                    }
                    ccollider.radius=val;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                0, 0.1f , false);
            height = new JSONStorableFloat("height",ccollider.height, (val) => 
                { 
                    if (val < 2*ccollider.radius) {
                      val = 2*ccollider.radius;
                      height.valNoCallback = val;  
                    } 
                    ccollider.height=val;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                0, 0.3f, false);
            x = new JSONStorableFloat("x",ccollider.center.x, (val) => 
                { 
                    center.x = val;
                    ccollider.transform.localPosition = center;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                -0.1f, 0.1f, false);
            y = new JSONStorableFloat("y",ccollider.center.y, (val) => 
                { 
                    center.y = val;
                    ccollider.transform.localPosition = center;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                -0.7f, 0.1f, false);
            z = new JSONStorableFloat("z",ccollider.center.z, (val) => 
                { 
                    center.z = val;
                    ccollider.transform.localPosition = center;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                -0.1f, 0.1f, false);
                
            rx = new JSONStorableFloat("rx", rotation.x, (val) => 
                {
                    rotation.x=val;
                    ccollider.transform.localEulerAngles = rotation;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                0f, 360f, true);                    
            ry = new JSONStorableFloat("ry",rotation.y, (val) => 
                { 
                    rotation.y=val;
                    ccollider.transform.localEulerAngles = rotation;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                0f, 360f, true); 
            rz = new JSONStorableFloat("rz",rotation.z, (val) => 
                { 
                    rotation.z=val;
                    ccollider.transform.localEulerAngles = rotation;
                    cgpu.UpdateData();
                    if (!isInterpolated) scr.interpFit();
                    UpdatePreview();
                },
                0f, 360f, true);                     
    
        }

        private Mesh makeTubeMesh()
        {
            Mesh m = new Mesh();
            Vector3 [] vertices = new Vector3[24];   
            for (int i=0; i<12; i++)
            {
                float r=0.5f;
                vertices[i] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f), -1, r*Mathf.Sin(i*2*Mathf.PI/12f));
            }
            for (int i=0; i<12; i++)
            {
                float r=0.5f;
                vertices[i+12] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f), 1, r*Mathf.Sin(i*2*Mathf.PI/12f));
            }

            int[] triangles = new int[72];
            for (int i=0;i<12; i++)
            {
                triangles[i*3] = i;
                triangles[i*3+1] = i+12;
                triangles[i*3+2] = (i+1) % 12;
            }
            for (int i=0;i<12; i++)
            {
                triangles[i*3+36] = (i+1) % 12;
                triangles[i*3+37] = i+12;
                triangles[i*3+38] = (i+1) % 12 + 12;
            }

            m.vertices = vertices;
            m.triangles = triangles;
            m.RecalculateNormals();
            m.RecalculateBounds();
            
            return m;
        }

        private Mesh makeHemisphereMesh()
        {
            Mesh m = new Mesh();
            Vector3 [] vertices = new Vector3[37];   
            for (int i=0; i<12; i++)
            {
                float h = 0;
                float r = 1f;
                vertices[i] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f)/2, h/2, r*Mathf.Sin(i*2*Mathf.PI/12f)/2);
            }
            for (int i=0; i<12; i++)
            {
                float h = Mathf.Sin(Mathf.PI/6);
                float r = Mathf.Cos(Mathf.PI/6);
                vertices[i+12] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f)/2, h/2, r*Mathf.Sin(i*2*Mathf.PI/12f)/2);
            }
            for (int i=0; i<12; i++)
            {
                float h = Mathf.Sin(Mathf.PI/3);
                float r = Mathf.Cos(Mathf.PI/3);
                vertices[i+24] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f)/2, h/2, r*Mathf.Sin(i*2*Mathf.PI/12f)/2);
            }
            vertices[36] = new Vector3(0,0.5f,0);


            int[] triangles = new int[180];
            for (int i=0;i<12; i++) // bottom row  /\ 0-35
            {
                triangles[i*3] = i;
                triangles[i*3+1] = i+12;
                triangles[i*3+2] = (i+1) % 12;
            }
            for (int i=0;i<12; i++) // bottom row \/ 36-71
            {
                triangles[i*3+36] = (i+1) % 12;
                triangles[i*3+37] = i+12;
                triangles[i*3+38] = (i+1) % 12 + 12;
            }

            for (int i=0;i<12; i++) // mid row /\ 72-107
            {
                triangles[i*3+72] = i+12;
                triangles[i*3+73] = i+24;
                triangles[i*3+74] = (i+1) % 12 + 12;
            }
            for (int i=0;i<12; i++) // mid row \/ 108-143
            {
                triangles[i*3+108] = (i+1) % 12 + 12;
                triangles[i*3+109] = i+24;
                triangles[i*3+110] = (i+1) % 12 + 24;
            }

            for (int i=0;i<12; i++) // top row /\ 144-179
            {
                triangles[i*3+144] = i+24;
                triangles[i*3+145] = 36;
                triangles[i*3+146] = (i+1) % 12 + 24;
            }



            m.vertices = vertices;
            m.triangles = triangles;
            m.RecalculateNormals();
            m.RecalculateBounds();
            
            return m;
        }


        public void createPreview()
        {
            showpreview=true;

            preview = new GameObject();
            MeshFilter mf = preview.AddComponent<MeshFilter>() as MeshFilter;
            preview.AddComponent<MeshRenderer>();
            mf.sharedMesh = makeTubeMesh();

            GameObject sphere1 = new GameObject();
            mf = sphere1.AddComponent<MeshFilter>() as MeshFilter;
            sphere1.AddComponent<MeshRenderer>();
            mf.sharedMesh = makeHemisphereMesh();
            sphere1.transform.position = preview.transform.position + Vector3.up;
            sphere1.transform.SetParent(preview.transform,true);


            GameObject sphere2 = new GameObject();
            mf = sphere2.AddComponent<MeshFilter>() as MeshFilter;
            sphere2.AddComponent<MeshRenderer>();
            mf.sharedMesh = makeHemisphereMesh();
            sphere2.transform.position = preview.transform.position - Vector3.up;
            sphere2.transform.Rotate(new Vector3(180,0,0));
            sphere2.transform.SetParent(preview.transform,true);



            preview.transform.SetParent(ccollider.transform, false);
            foreach (var c in preview.GetComponentsInChildren<Collider>())
            {
                c.enabled = false;
                UnityEngine.Object.Destroy(c);
            }

            // setInactive();
            UpdatePreview();

        }
        public void destroyPreview()
        {
            if (showpreview) 
            {
                UnityEngine.Object.Destroy(preview);
                showpreview = false;
            }


        }

        public void UpdatePreview()
        {
            // SuperController.LogMessage($"updating preview for {name}");
            if (showpreview) {
              var allPreviewRenderers = preview.GetComponentsInChildren<Renderer>(); 
                // SuperController.LogMessage($"renderers : {previewrenderers.Length} ");
                // var previewRenderer = preview.GetComponent<Renderer>();
              foreach (var previewRenderer in allPreviewRenderers)
              {
                var material = previewRenderer.material;
                if (seethrough)
                {
                    material.shader = Shader.Find("Battlehub/RTGizmos/Handles");
                    material.SetFloat("_Offset", 1f);
                    material.SetFloat("_MinAlpha", 1f);
                    previewRenderer.material = material;
                }
                else
                {
                    material.shader = Shader.Find("Standard");
                    material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.SrcAlpha);
                    material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
                    material.SetInt("_ZWrite", 0);
                    material.DisableKeyword("_ALPHATEST_ON");
                    material.EnableKeyword("_ALPHABLEND_ON");
                    material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
                    material.renderQueue = 3000;
                    previewRenderer.material = material;
                }

                if (gui) {
                    var color = Color.magenta;//previewRenderer.material.color;
                    color.a = seethrough ? Mathf.Clamp01(seethroughalpha * 10) : Mathf.Clamp01(occludedalpha*2);
                    previewRenderer.material.color = color;
                } 
                else
                {
                    var color = interp_n1!=null && interp_n2!=null ? Color.yellow : Color.green;//previewRenderer.material.color;
                    color.a = seethrough ? seethroughalpha : occludedalpha;
                    previewRenderer.material.color = color;
                    // if (previewRenderer.transform == preview.transform.GetChild(0)) {
                    //   color = Color.blue;
                    //   color.a = seethrough ? seethroughalpha : occludedalpha;
                    //   previewRenderer.material.color = color;

                    // }
                }
              }

              float size1 = cgpu.RadiusA * 2;
              float size2 = cgpu.RadiusB * 2;
              float size = 0.5f*(size1+size2);
              float height = ccollider.height / 2 - size1 / 4 - size2 / 4;
              if (height==0f) height = 0.00001f;
              if (size==0f) size = 0.00001f;
              preview.transform.localScale = new Vector3(size, height, size);
              preview.transform.GetChild(0).localScale = new Vector3(size1/size, size1/height, size1/size);
              preview.transform.GetChild(1).localScale = new Vector3(size2/size, size2/height, size2/size);
              if (ccollider.direction == 0)
                  preview.transform.localRotation = Quaternion.AngleAxis(90, Vector3.forward);
              else if (ccollider.direction == 2)
                  preview.transform.localRotation = Quaternion.AngleAxis(90, Vector3.right);
            }
        }

        public void createGUI(MVRScript scr)
        {
            s_radius = scr.CreateSlider(radius);
            s_height = scr.CreateSlider(height);
            s_x = scr.CreateSlider(x);
            s_y = scr.CreateSlider(y);
            s_z = scr.CreateSlider(z);
            s_radius.valueFormat="F8";
            s_height.valueFormat="F8";
            s_x.valueFormat="F8";
            s_y.valueFormat="F8";
            s_z.valueFormat="F8";
            s_rx = scr.CreateSlider(rx,true);
            s_ry = scr.CreateSlider(ry,true);
            s_rz = scr.CreateSlider(rz,true);

            // setActive();
            gui = true;
            UpdatePreview();
        }

        public void destroyGUI(MVRScript scr)
        {
            if (!gui) return;
            scr.RemoveSlider(s_radius);
            scr.RemoveSlider(s_height);
            scr.RemoveSlider(s_x);
            scr.RemoveSlider(s_y);
            scr.RemoveSlider(s_z);
            scr.RemoveSlider(s_rx);
            scr.RemoveSlider(s_ry);
            scr.RemoveSlider(s_rz);
            // setInactive();
            gui = false;
            UpdatePreview();
        }


        public void mirrorData(FakeCollider source) 
        {
            SetTransform(source.radius.val, source.height.val, -source.x.val, source.y.val, source.z.val, source.rx.val, 360-source.ry.val, 360-source.rz.val);
        }

        public void SetTransform(float rad, float h, float xx, float yy, float zz, float rxx, float ryy, float rzz)
        {
            radius.val = rad;
            height.val = h;
            x.val = xx;
            y.val = yy;
            z.val = zz;
            rx.val = rxx;
            ry.val = ryy;
            rz.val = rzz;
        }

        public void SetTransformNoCallback(float rad, float h, float xx, float yy, float zz, float rxx, float ryy, float rzz)
        {
            ccollider.radius = rad;
            ccollider.height = h;
            center.x = xx;
            center.y = yy;
            center.z = zz;
            ccollider.transform.localPosition = center;
            rotation.x = rxx;
            rotation.y = ryy;
            rotation.z = rzz;            
            ccollider.transform.localEulerAngles = rotation;


        }


        public void Autofit(Vector3[] verts)
        {
            if (AutofitInstructions.instructions.ContainsKey(name)) {
                AutofitSingle instr = AutofitInstructions.instructions[name] as AutofitSingle;

                autofitData data = instr.calculate(verts, anchor.transform);

                Vector3 pos = data.center;
                float r = data.radius;
                float h = data.height;
                Vector3 rot = data.angles;
                float offset =  0f;
                float addh = 0f;
                SetTransform(r, h+addh, pos.x, pos.y-offset, pos.z, rot.x, rot.y, rot.z);
            }

        }

        public void InterpolatedFit( List<FakeCollider> list )
        {
            if (interp_n1==null || interp_n2==null || interp_n1==interp_n2) return;
            // SuperController.LogMessage($"interpfitting {name} ...");
            
            Vector3 pos_w = (interp_n1.anchor.transform.TransformPoint(interp_n1.center) + 
                             interp_n2.anchor.transform.TransformPoint(interp_n2.center))/2;
            Vector3 rot_w = Quaternion.Slerp(
                                Quaternion.Euler(interp_n1.rotation),
                                Quaternion.Euler(interp_n2.rotation)
                                , 0.5f).eulerAngles;
            Vector3 pos = anchor.transform.InverseTransformPoint(pos_w);
            Vector3 rot = rot_w;
            SetTransformNoCallback(
                (interp_n1.radius.val+interp_n2.radius.val)/2,
                (interp_n1.height.val+interp_n2.height.val)/2,
                pos.x, pos.y, pos.z,
                rot.x, rot.y, rot.z
            );
            cgpu.UpdateData();
            // cgpu.RadiusA = interp_n1.radius.val;
            // cgpu.RadiusB = interp_n2.radius.val;
            UpdatePreview();
            // SuperController.LogMessage($"interpfitting done");

        }

        public void InterpolatedFitNew( List<FakeCollider> list )
        {
            if (interp_n1==null || interp_n2==null || interp_n1==interp_n2) return;
            // SuperController.LogMessage($"interpfitting {name} ...");

            Vector3 point_w1 = isInterpA ? interp_n1.cgpu.WorldA : interp_n1.cgpu.WorldB;
            Vector3 point_w2 = isInterpA ? interp_n2.cgpu.WorldA : interp_n2.cgpu.WorldB;
            Vector3 pos_w = (point_w1 + point_w2)/2;
            Vector3 pos = anchor.transform.InverseTransformPoint(pos_w);

            Quaternion rot_w = Quaternion.LookRotation(pos_w - point_w2);

            Quaternion rot = Quaternion.Inverse(anchor.transform.rotation)*rot_w*Quaternion.Euler(90,0,0);
            Vector3 rotv = rot.eulerAngles;

            SetTransformNoCallback(
                (interp_n1.radius.val+interp_n2.radius.val)/2,
                (interp_n2.cgpu.WorldA - interp_n1.cgpu.WorldA).magnitude,
                pos.x, pos.y, pos.z,
                rotv.x, rotv.y, rotv.z
            );
            cgpu.UpdateData();
            cgpu.WorldA = point_w1;
            cgpu.WorldB = point_w2;
            cgpu.RadiusA = interp_n1.cgpu.RadiusA; // radius A and B should be the same for non interp colliders
            cgpu.RadiusB = interp_n2.cgpu.RadiusA;
            UpdatePreview();
            // SuperController.LogMessage($"new interpfitting done");

        }

        public void Update() {  
        }

        public void FcolDestroy() {
            UnityEngine.Object.Destroy(parent.gameObject,0.5f);
            destroyPreview();
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            jc["radius"].AsFloat = radius.val;
            jc["height"].AsFloat = height.val;
            jc["x"].AsFloat = x.val;
            jc["y"].AsFloat = y.val;
            jc["z"].AsFloat = z.val;
            jc["rx"].AsFloat = rx.val;
            jc["ry"].AsFloat = ry.val;
            jc["rz"].AsFloat = rz.val;
            return jc;
        }

        public void RestoreFromJSON(JSONClass jc, bool correctoldversion = false)
        {
            radius.val = jc["radius"].AsFloat;
            height.val = jc["height"].AsFloat;
            radius.val = jc["radius"].AsFloat; //twice to account for limiting
            height.val = jc["height"].AsFloat;
            x.val = jc["x"].AsFloat;
            y.val = jc["y"].AsFloat;
            if (correctoldversion) y.val += 0.06f;
            z.val = jc["z"].AsFloat;
            rx.val = jc["rx"].AsFloat;
            ry.val = jc["ry"].AsFloat;
            rz.val = jc["rz"].AsFloat;
        }
        
    }