using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using SimpleJSON;
using GPUTools.Physics.Scripts.Behaviours;
using System.Linq;
using Battlehub.UIControls;

public class UYCol : MVRScript
{
    List<FakeCollider> fakecolliders;
    List<FakeCollider> interpolatedcolliders;
    List<FakeCollider> interpolatedcollidersNew;

    JSONStorableStringChooser fakecolliderchooser;
    JSONClass manualJC;
    string _savePath = "Saves\\UYFakeColliders";
    string _saveExt = "uy";

    UIDynamicToggle toggleshow;
    UIDynamicToggle toggleseethrough;
    UIDynamicSlider slideralpha;

    UIDynamicTextField labelhelp;
    UIDynamicButton buttonload;
    UIDynamicButton buttonsave;
    UIDynamicPopup popupchooser;

    UIDynamicButton buttoncustomize;
    UIDynamicButton buttonLR;
    UIDynamicButton buttonRL;
    UIDynamicButton buttonFit;
    bool LRconfirm;
    bool RLconfirm;
    bool fitconfirm;
    // Vector3[] autoverts;

    JSONStorableBool boolshow;
    JSONStorableBool boolseethrough;
    JSONStorableFloat alpha;
    

    UIDynamicToggle useLegs;
    UIDynamicToggle useArms;
    UIDynamicToggle useHip;
    UIDynamicToggle useTorso;

    UIDynamicToggle toggleinterp;
    UIDynamicToggle toggleinterpNew;    
    UIDynamicToggle toggleshowinterp;

    UIDynamicToggle toggleConvertExistingLegs;
    UIDynamicToggle toggleConvertExistingArms;
    UIDynamicToggle toggleConvertExistingHip;
    UIDynamicToggle toggleConvertExistingTorso;
    UIDynamicToggle toggleAlsoCollideLegs;
    UIDynamicToggle toggleAlsoCollideArms;
    UIDynamicToggle toggleAlsoCollideHip;
    UIDynamicToggle toggleAlsoCollideTorso;
    List<UIDynamic> spacers;


    JSONStorableBool boolLegs;
    JSONStorableBool boolArms;
    JSONStorableBool boolHip;
    JSONStorableBool boolTorso;
    JSONStorableBool boolConvertExistingLegs;
    JSONStorableBool boolConvertExistingArms;
    JSONStorableBool boolConvertExistingHip;
    JSONStorableBool boolConvertExistingTorso;

    JSONStorableBool boolAlsoCollideLegs;
    JSONStorableBool boolAlsoCollideArms;
    JSONStorableBool boolAlsoCollideHip;
    JSONStorableBool boolAlsoCollideTorso;

    List<string> legRigidbodies = new List<string>{"rShin", "rThigh", "lShin", "lThigh", "lFoot", "rFoot"};
    List<string> armRigidbodies = new List<string>{"lShldr", "lForearm", "rShldr", "rForearm"};
    List<string> hipRigidbodies = new List<string>{"pelvis", "hip"};
    List<string> torsoRigidbodies = new List<string>{"abdomen", "abdomen2", "chest", "lPectoral", "rPectoral"};

    JSONStorableBool boolinterp;
    JSONStorableBool boolinterpNew;
    JSONStorableBool boolshowinterp;

    UIDynamicButton buttoncustomdone;
    UIDynamicTextField labelhelp2;

    List<CapsuleLineSphereCollider> modified;

    // List<string> interpNames;
    List<string> fcolNames;
    // List<Atom> atoms;
    private bool legs_created=false;
    private bool arms_created=false;
    private bool hip_created=false;
    private bool torso_created=false;
    private bool interp_created=false;
    private bool is_loading=true;
    private bool was_loading=true;

    private void createCapsule(string anchorname, string fcolname, List<string> ignoreRBs, string interp1="", string interp2="", bool interpnew = false)
    {
        string atomname = fcolname;


        if (interp1!="" && interp2!="") 
        {
            FakeCollider fcol1 = null;
            FakeCollider fcol2 = null;
            foreach (FakeCollider f in fakecolliders)
            {
                if (f.name == interp1) fcol1=f;
                if (f.name == interp2) fcol2=f;
            }
            
            if (interpnew) {
              interpolatedcollidersNew.Add(new FakeCollider(this,fcolname+"_A", anchorname, ignoreRBs, fcol1, fcol2, true));
              interpolatedcollidersNew.Add(new FakeCollider(this,fcolname+"_B", anchorname, ignoreRBs, fcol1, fcol2, false));
            }
            else interpolatedcolliders.Add(new FakeCollider(this,fcolname, anchorname, ignoreRBs, fcol1, fcol2));
            // interpNames.Add(fcolname);
        }
        else
        {
            fakecolliders.Add(new FakeCollider(this,fcolname, anchorname, ignoreRBs));
            fcolNames.Add(fcolname);
        }
    }

    private void createLegs()
    {
        legs_created = false;
        for (int i=0; i<8; i++)
        {
            createCapsule("lThigh", $"lthigh{i}",
                            new List<string>{"lShin", "lThigh", "rThigh", "abdomen",
                                "JointB1Male","JointB2Male","JointAlMale","JointArMale",
                                "pelvis", "Gen","Testes"});
            createCapsule("rThigh", $"rthigh{i}",
                            new List<string>{"rShin", "rThigh", "lThigh", "abdomen",
                                "JointB1Male","JointB2Male","JointAlMale","JointArMale",
                                "pelvis", "Gen","Testes"});
        }
        for (int i=0; i<9; i++)
        {
            createCapsule( "lShin", $"lshin{i}", 
                            new List<string>{"lShin", "lThigh","lFoot"});
            createCapsule("rShin", $"rshin{i}",
                            new List<string>{"rShin", "rThigh","rFoot"});

        }
        legs_created = true;
        fcolNames.Sort();
    }

    private void createArms()
    {
        arms_created = false;
        createCapsule("lCollar", $"lcollar0",
                      new List<string>{"lShldr", "chest", "lPectoral"});
        createCapsule("rCollar", $"rcollar0", 
                      new List<string>{"rShldr", "chest", "rPectoral"});

        for (int i=0; i<6; i++)
        {
            createCapsule( "lShldr", $"lshould{i}",
                      new List<string>{"lForearm", "lShldr", "chest", "lPectoral"});
            createCapsule("rShldr", $"rshould{i}",
                      new List<string>{"rForearm", "rShldr", "chest", "rPectoral"});
        }        

        arms_created = true;    
        fcolNames.Sort();
    }

    private void createHip()
    {
        hip_created = false;
        for (int i=0; i<1; i++)
        {
            createCapsule("pelvis", $"lpelvis{i}",
                      new List<string>{"abdomen", "abdomen2","pelvis", "chest", "lThigh",
                      "JointB1Male","JointB2Male","JointAlMale","JointArMale", "Gen1","Testes"});            
            createCapsule("pelvis", $"rpelvis{i}",
                      new List<string>{"abdomen", "abdomen2","pelvis", "chest", "rThigh",
                      "JointB1Male","JointB2Male","JointAlMale","JointArMale", "Gen1","Testes"});            

        }
        for (int i=0; i<2; i++)
        {
            createCapsule("hip", $"mhip{i}",
                      new List<string>{"abdomen", "abdomen2","pelvis", "chest", "rThigh", "lThigh",
                      "JointB1Male","JointB2Male","JointAlMale","JointArMale", "Gen1","Testes"});

        }
        for (int i=0; i<1; i++)
        {
            createCapsule("abdomen", $"mabdomenA{i}",
                      new List<string>{"abdomen", "abdomen2","pelvis", "chest", "rThigh", "lThigh",
                      "JointB1Male","JointB2Male","JointAlMale","JointArMale", "Gen1","Testes"});

        }        
        hip_created = true;
        fcolNames.Sort();

    }

    private void createTorso()
    {
        torso_created = false;
        for (int i=0; i<1; i++)
        {
            createCapsule("lPectoral", $"lbreast{i}",
                      new List<string>{"lShldr", "chest", "lPectoral", "rPectoral",
                      "abdomen2", "abdomen"});
            createCapsule("rPectoral", $"rbreast{i}",
                      new List<string>{"rShldr", "chest", "lPectoral", "rPectoral",
                      "abdomen2", "abdomen"});

        }
        createCapsule("lCollar", $"lcollar1",
                      new List<string>{"lShldr", "chest", "lPectoral", "rPectoral",
                       "abdomen2", "abdomen", "neck"});
        createCapsule("rCollar", $"rcollar1",
                      new List<string>{"lShldr", "chest", "lPectoral", "rPectoral",
                      "abdomen2", "abdomen", "neck"});

        for (int i=0; i<3; i++)
        {
            createCapsule("abdomen2", $"mabdomenB{i}",
                      new List<string>{"abdomen", "chest", "abdomen2", "pelvis", "hip",
                      "lShldr","rShldr", "lPectoral", "rPectoral", "lThigh", "rThigh"});
            
        }
        for (int i=0; i<3; i++)
        {
            createCapsule("chest", $"mchest{i}",
                      new List<string>{"abdomen", "chest", "abdomen2", "pelvis", "hip",
                      "lShldr","rShldr","lPectoral", "rPectoral"});

        } 
        for (int i=0; i<3; i++)
        {
            createCapsule("chest", $"lchest{i}",
                      new List<string>{"abdomen", "chest", "abdomen2", "pelvis", "hip",
                      "lPectoral", "rPectoral", "lShldr" });
            createCapsule("chest", $"rchest{i}",
                      new List<string>{"abdomen", "chest", "abdomen2", "pelvis", "hip",
                      "lPectoral", "rPectoral", "rShldr" });
        }                
        torso_created = true;
        fcolNames.Sort();
    }


    private void createInterpolated()
    {
        interp_created = false;
        // SuperController.LogMessage($"IN create interpolated");

        if (boolLegs.val) 
        {
            for (int i=0; i<8; i++)
            {
                createCapsule("lShin", $"lshinI{i}", null, $"lshin{i}", $"lshin{i+1}");
                createCapsule("rShin", $"rshinI{i}", null, $"rshin{i}", $"rshin{i+1}");
            }
            for (int i=0; i<4; i++)
            {
                createCapsule("lThigh", $"lthighI{i}", null, $"lthigh{i}", $"lthigh{i+1}"); 
                createCapsule("rThigh", $"rthighI{i}", null, $"rthigh{i}", $"rthigh{i+1}"); 
            }
            // special cases thigh: t0 ... t3 t34 t4 t47 t7 t75 t5 t56 t6
            createCapsule("lThigh", $"lthighI4", null, $"lthigh4", $"lthigh7"); 
            createCapsule("lThigh", $"lthighI5", null, $"lthigh7", $"lthigh5"); 
            createCapsule("lThigh", $"lthighI6", null, $"lthigh5", $"lthigh6"); 
            
            // special cases thigh: t0 ... t3 t34 t4 t47 t7 t75 t5 t56 t6
            createCapsule("rThigh", $"rthighI4", null, $"rthigh4", $"rthigh7"); 
            createCapsule("rThigh", $"rthighI5", null, $"rthigh7", $"rthigh5"); 
            createCapsule("rThigh", $"rthighI6", null, $"rthigh5", $"rthigh6"); 
        }

        if (boolArms.val)
        {
            for (int i=0; i<5; i++)
            {
                createCapsule("lShldr", $"lshouldI{i}", null, $"lshould{i}", $"lshould{i+1}");
                createCapsule("rShldr", $"rshouldI{i}", null, $"rshould{i}", $"rshould{i+1}");
            }
        }
        if (boolHip.val)
        {
            createCapsule("hip", "mhipI0", null, "mhip0", "mabdomenA0");
        }

        if (boolTorso.val)
        {    
            createCapsule("abdomen2", "mabdomenBI0", null, "mabdomenB0", "mabdomenB1");
            createCapsule("abdomen2", "mabdomenBI1", null, "mabdomenB1", "mchest0");
            createCapsule("chest", "mchestI0", null, "mchest0", "mchest1");
            createCapsule("chest", "mchestI1", null, "mchest1", "mchest2");
        }
        if (boolHip.val && boolTorso.val) {
            createCapsule("abdomen", "mabdomenAI0", null, "mabdomenA0", "mabdomenB0");
        }
        // SuperController.LogMessage($"created interpolated");
        interp_created = true;
        interpFit();

    }

    private void createInterpolatedNew()
    {
        interp_created = false;
        // SuperController.LogMessage($"IN create interpolated");

        if (boolLegs.val) 
        {
            for (int i=0; i<8; i++)
            {
                createCapsule("lShin", $"lshinI{i}", null, $"lshin{i}", $"lshin{i+1}", true);
                createCapsule("rShin", $"rshinI{i}", null, $"rshin{i}", $"rshin{i+1}", true);
            }
            for (int i=0; i<4; i++)
            {
                createCapsule("lThigh", $"lthighI{i}", null, $"lthigh{i}", $"lthigh{i+1}", true); 
                createCapsule("rThigh", $"rthighI{i}", null, $"rthigh{i}", $"rthigh{i+1}", true); 
            }
            // special cases thigh: t0 ... t3 t34 t4 t47 t7 t75 t5 t56 t6
            createCapsule("lThigh", $"lthighI4", null, $"lthigh4", $"lthigh7"); 
            createCapsule("lThigh", $"lthighI5", null, $"lthigh7", $"lthigh5"); 
            createCapsule("lThigh", $"lthighI6", null, $"lthigh5", $"lthigh6"); 
            
            // special cases thigh: t0 ... t3 t34 t4 t47 t7 t75 t5 t56 t6
            createCapsule("rThigh", $"rthighI4", null, $"rthigh4", $"rthigh7"); 
            createCapsule("rThigh", $"rthighI5", null, $"rthigh7", $"rthigh5"); 
            createCapsule("rThigh", $"rthighI6", null, $"rthigh5", $"rthigh6"); 
        }

        if (boolArms.val)
        {
            for (int i=0; i<5; i++)
            {
                createCapsule("lShldr", $"lshouldI{i}", null, $"lshould{i}", $"lshould{i+1}", true);
                createCapsule("rShldr", $"rshouldI{i}", null, $"rshould{i}", $"rshould{i+1}", true);
            }
        }

        if (boolArms.val && boolTorso.val) {
                createCapsule("lCollar", $"lCollarI0", null, $"lcollar1", $"lcollar0", true);
                createCapsule("rCollar", $"rCollarI0", null, $"rcollar1", $"rcollar0", true);

        }
        
        if (boolHip.val)
        {
            createCapsule("hip", "mhipI0", null, "mhip0", "mabdomenA0", true);
        }

        if (boolTorso.val)
        {    
            createCapsule("abdomen2", "mabdomenBI0", null, "mabdomenB0", "mabdomenB1", true);
            createCapsule("abdomen2", "mabdomenBI1", null, "mabdomenB1", "mchest0", true);
            createCapsule("chest", "mchestI0", null, "mchest0", "mchest1", true);
            createCapsule("chest", "mchestI1", null, "mchest1", "mchest2", true);
        }
        if (boolHip.val && boolTorso.val) {
            createCapsule("abdomen", "mabdomenAI0", null, "mabdomenA0", "mabdomenB0", true);
        }
        // SuperController.LogMessage($"created interpolated");
        interp_created = true;
        interpFit();

    }


    private void removeLegs()
    {
        for (int i=0; i<9; i++)
        {
            removeAtomFull($"lshin{i}");
            removeAtomFull($"rshin{i}");
        }
        for (int i=0; i<8; i++)
        {
           removeAtomFull($"lthigh{i}");
           removeAtomFull($"rthigh{i}");
        }        
    }
    
    private void removeArms()
    {
        removeAtomFull("lcollar0");
        removeAtomFull("rcollar0");
        for (int i=0; i<6; i++)
        {
            removeAtomFull($"lshould{i}");
            removeAtomFull($"rshould{i}");
        }  
    }
    
    private void removeHip()
    {
        for (int i=0; i<1; i++)
        {
           removeAtomFull($"lpelvis{i}");
           removeAtomFull($"rpelvis{i}");
        }
        for (int i=0; i<2; i++)
        {
            removeAtomFull($"mhip{i}");
        }
        for (int i=0; i<1; i++)
        {
            removeAtomFull($"mabdomenA{i}");
        }          
    }

    private void removeTorso()
    {
        
        for (int i=0; i<1; i++)
        {
            removeAtomFull($"lbreast{i}");
            removeAtomFull($"rbreast{i}");
        }
        removeAtomFull($"lcollar1");
        removeAtomFull($"rcollar1");
        for (int i=0; i<3; i++)
        {
            removeAtomFull($"mabdomenB{i}");
        }
        for (int i=0; i<3; i++)
        {
            removeAtomFull($"mchest{i}");
        }
        for (int i=0; i<3; i++)
        {
            removeAtomFull($"lchest{i}");
            removeAtomFull($"rchest{i}");
        } 
    }

    private void removeAtomFull(string aid)
    {
        if (fcolNames.Contains(aid)) 
        {
            fcolNames.Remove(aid);
        }
        List<FakeCollider> toremove = new List<FakeCollider>();
        foreach(FakeCollider fcol in fakecolliders)
        {
            if (fcol.name==aid)
            {
               toremove.Add(fcol);
            }
        }
        // SuperController.LogMessage("hmm1");
        foreach(FakeCollider fcol in toremove)
        {   
            fcol.FcolDestroy();
            fakecolliders.Remove(fcol);
        }
        // SuperController.LogMessage("hmm2");
        
    }
    private void removeFcolAtoms()
    {

        fcolNames.Clear();
        foreach (FakeCollider fcol in fakecolliders)
        {
            fcol.FcolDestroy();
        }
        fakecolliders.Clear();

    }
    private void removeInterpolated()
    {
        // interpNames.Clear();
        foreach (FakeCollider fcol in interpolatedcolliders)
        {
            fcol.FcolDestroy();
        }        
        interpolatedcolliders.Clear();
    }

    private void removeInterpolatedNew()
    {
        // interpNames.Clear();
        foreach (FakeCollider fcol in interpolatedcollidersNew)
        {
            fcol.FcolDestroy();
        }        
        interpolatedcollidersNew.Clear();
    }


    public void AttemptAddGPUToExisting (List<string> rigidbodyfilter)
    {
      var allcolliders = containingAtom.GetComponentsInChildren<CapsuleCollider>().Where(
        (c)=> { return rigidbodyfilter.Any((filter) => c.attachedRigidbody.name.Contains(filter)); }
      ); 
      var allgpucolliders = containingAtom.GetComponentsInChildren<CapsuleLineSphereCollider>();
      int count=0;
      foreach( CapsuleCollider col in allcolliders ){

        foreach(CapsuleLineSphereCollider gpucol in allgpucolliders) {
          if (gpucol.capsuleCollider == col) {
            // SuperController.LogMessage($"collider {col.name} has gpu counterpart {gpucol.name}");
            continue;
          } 
        }
        if (col.isTrigger==false) {
          // SuperController.LogMessage($"collider {col.name} has no gpu counterpart");
          var newgpucol = col.gameObject.AddComponent<CapsuleLineSphereCollider>();
          newgpucol.capsuleCollider = col;
          newgpucol.UpdateData();
          modified.Add(newgpucol);
          count++;
        }

      } 
      SuperController.LogMessage($" plain colliders found {allcolliders.Count()} : gpu colliders added {count}");
    }

    public void RestoreModifiedExisting(List<string> rigidbodyfilter) {
      int count = 0;
      foreach(CapsuleLineSphereCollider clsc in modified) {
        if (rigidbodyfilter.Any((filter)=> clsc.capsuleCollider.attachedRigidbody.name.Contains(filter))) {
          Destroy(clsc, 0.5f);
          count++;
        }
      }
      SuperController.LogMessage($"GPU colliders removed {count}");
    }

    public void RemoveOldAtoms() {
      var allatoms = SuperController.singleton.GetAtoms();
      foreach (var oldatom in allatoms) {
        if (oldatom.name.StartsWith("uycol_")) {
          SuperController.singleton.RemoveAtom(oldatom);
        }
      }
    }



    public override void Init()
    {
        // SuperController.LogMessage("INSIDE INIT");
        RemoveOldAtoms();
        modified = new List<CapsuleLineSphereCollider>();

        fcolNames = new List<string>();
        fakecolliders = new List<FakeCollider>();
        interpolatedcolliders = new List<FakeCollider>();
        interpolatedcollidersNew = new List<FakeCollider>();

        boolshow = new JSONStorableBool("Show Colliders", false, (val) => {
            if (val) {
                foreach( FakeCollider fcol in fakecolliders) {
                    fcol.createPreview();
                }
            } else {
                foreach( FakeCollider fcol in fakecolliders) {
                    fcol.destroyPreview();
                }
            }
            RefreshUI(fakecolliderchooser.val);

        });

        boolseethrough = new JSONStorableBool("See-through Colliders", true, (val) => {
            foreach( FakeCollider fcol in fakecolliders) {
                fcol.seethrough = val;
                fcol.UpdatePreview();
            }
            RefreshUI(fakecolliderchooser.val);

        });

        alpha = new JSONStorableFloat("Opacity",0.05f, (val) =>
        {
            FakeCollider.seethroughalpha = val;
            FakeCollider.occludedalpha = Mathf.Clamp01(val*4);
            foreach (FakeCollider fcol in fakecolliders) 
            {
                fcol.UpdatePreview();
            }
            foreach (FakeCollider fcol in interpolatedcolliders) 
            {
                fcol.UpdatePreview();
            }
        },
        0f, 1f, true );

        boolLegs = new JSONStorableBool("Use Legs", true, (val) => {
            // SuperController.LogMessage($"inside legs callback {val}");
            if (val) {
                createLegs(); 
            } else {
                removeLegs();
            }
        });    


        boolArms = new JSONStorableBool("Use Arms", true, (val) => {
            // SuperController.LogMessage($"inside arms callback {val}");
            if (val) {
                createArms(); 
            } else {
                removeArms();
            }
        });


        boolHip = new JSONStorableBool("Use Hip", false, (val) => {
            // SuperController.LogMessage($"inside hip callback {val}");
            if (val) {
                createHip(); 
            } else {
                removeHip();
            }
        });

        boolTorso = new JSONStorableBool("Use Torso", true, (val) => {
            // SuperController.LogMessage($"inside torso callback {val}");
            if (val) {
                createTorso(); 
            } else {
                removeTorso();
            }
        });

        boolinterp = new JSONStorableBool("Use Interpolated Colliders", false, (val) => {
            // SuperController.LogMessage($"inside interp callback {val}");
            if (val) {
                createInterpolated(); // use only what is shown?
            } else {
                removeInterpolated();
                // interpNames.Clear();
                // interpolatedcolliders.Clear();
            }

        });
        boolinterpNew = new JSONStorableBool("Use NEW Interpolated Colliders", false, (val) => {
            // SuperController.LogMessage($"inside interp new callback {val}");
            if (val) {
                createInterpolatedNew(); // use only what is shown?
            } else {
                removeInterpolatedNew();
                // interpNames.Clear();
                // interpolatedcolliders.Clear();
            }

        });

        boolshowinterp = new JSONStorableBool("Show Interpolated Colliders", false, (val) => {
            if (val) {
                foreach( FakeCollider fcol in interpolatedcolliders) {
                    fcol.createPreview();
                }
                foreach( FakeCollider fcol in interpolatedcollidersNew) {
                    fcol.createPreview();
                }                
            } else {
                foreach( FakeCollider fcol in interpolatedcolliders) {
                    fcol.destroyPreview();
                }
                foreach( FakeCollider fcol in interpolatedcollidersNew) {
                    fcol.destroyPreview();
                }                
            }
            // RefreshUI(fakecolliderchooser.val);

        });

        boolConvertExistingLegs = new JSONStorableBool("Builtin Leg Colliders to Cloth", false, (val) => {
            if (val) {
              AttemptAddGPUToExisting(legRigidbodies);
            } else {
              RestoreModifiedExisting(legRigidbodies);
            }
            // RefreshUI(fakecolliderchooser.val);

        });

        boolConvertExistingArms = new JSONStorableBool("Builtin Arm Colliders To Cloth", false, (val) => {
            // SuperController.LogMessage("inside callback exist arms");
            if (val) {
              AttemptAddGPUToExisting(armRigidbodies);
            } else {
              RestoreModifiedExisting(armRigidbodies);
            }
            // RefreshUI(fakecolliderchooser.val);

        });

        boolConvertExistingHip = new JSONStorableBool("Builtin Hip Colliders To Cloth", true, (val) => {
            if (val) {
              AttemptAddGPUToExisting(hipRigidbodies);
            } else {
              RestoreModifiedExisting(hipRigidbodies);
            }
            // RefreshUI(fakecolliderchooser.val);

        });
        AttemptAddGPUToExisting(hipRigidbodies); // callback wont be called on init, ???


        boolConvertExistingTorso = new JSONStorableBool("Builtin Torso Colliders To Cloth", false, (val) => {
            if (val) {
              AttemptAddGPUToExisting(torsoRigidbodies);
            } else {
              RestoreModifiedExisting(torsoRigidbodies);
            }
            // RefreshUI(fakecolliderchooser.val);

        });

        boolAlsoCollideLegs = new JSONStorableBool("Legs Collide with Other Things", false, (val) => {
            if (val) {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(legRigidbodies, false));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(legRigidbodies, false));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(legRigidbodies, false));
            } else {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(legRigidbodies, true));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(legRigidbodies, true));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(legRigidbodies, true));
            }
            // RefreshUI(fakecolliderchooser.val);
        });

        boolAlsoCollideArms = new JSONStorableBool("Arms Collide with Other Things", false, (val) => {
            if (val) {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(armRigidbodies, false));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(armRigidbodies, false));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(armRigidbodies, false));
            } else {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(armRigidbodies, true));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(armRigidbodies, true));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(armRigidbodies, true));
            }
            // RefreshUI(fakecolliderchooser.val);
        });


        boolAlsoCollideHip = new JSONStorableBool("Hip Collides with Other Things", false, (val) => {
            if (val) {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(hipRigidbodies, false));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(hipRigidbodies, false));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(hipRigidbodies, false));
            } else {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(hipRigidbodies, true));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(hipRigidbodies, true));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(hipRigidbodies, true));
            }
            // RefreshUI(fakecolliderchooser.val);
        });
        
        boolAlsoCollideTorso = new JSONStorableBool("Torso Collides with Other Things", false, (val) => {
            if (val) {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(torsoRigidbodies, false));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(torsoRigidbodies, false));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(torsoRigidbodies, false));
            } else {
              fakecolliders.ForEach((fc)=> fc.SetTrigger(torsoRigidbodies, true));
              interpolatedcolliders.ForEach((fc)=> fc.SetTrigger(torsoRigidbodies, true));
              interpolatedcollidersNew.ForEach((fc)=> fc.SetTrigger(torsoRigidbodies, true));
            }
            // RefreshUI(fakecolliderchooser.val);
        });

        spacers = new List<UIDynamic>();

        RegisterBool(boolLegs);
        RegisterBool(boolArms);
        RegisterBool(boolHip);
        RegisterBool(boolTorso);
        RegisterBool(boolinterp);
        RegisterBool(boolinterpNew);
        RegisterBool(boolConvertExistingLegs);
        RegisterBool(boolConvertExistingArms);
        RegisterBool(boolConvertExistingHip);
        RegisterBool(boolConvertExistingTorso);
        RegisterBool(boolAlsoCollideLegs);
        RegisterBool(boolAlsoCollideArms);
        RegisterBool(boolAlsoCollideHip);
        RegisterBool(boolAlsoCollideTorso);

        // showGUI1();
        CreateAll();
        // showGUI1();


    }        



    public void CreateAll(bool isoldversion=false)
    {
        // SuperController.LogMessage("SCENE LOADED starting to create atoms");
        // StartCoroutine(createAllAtoms());
        if (boolLegs.val)
            createLegs(); 
        if (boolArms.val)
            createArms(); 
        if (boolHip.val)
            createHip(); 
        if (boolTorso.val)
            createTorso(); 
        if (boolinterp.val)
            createInterpolated();
        manualRestore(null, isoldversion);
        showGUI1();
    }

    public UIDynamicPopup CreateAutoPopup(JSONStorableStringChooser jssc, bool rightSide = false)
    {
        UIDynamicPopup popup = CreateFilterablePopup(jssc, rightSide);
        popup.popup.labelText.alignment = TextAnchor.UpperCenter;
        popup.popup.labelText.GetComponent<RectTransform>().anchorMax = new Vector2(0, 0.89f);
        
        {
            var btn = Instantiate(manager.configurableButtonPrefab);
            btn.SetParent(popup.transform, false);
            Destroy(btn.GetComponent<LayoutElement>());
            btn.GetComponent<UIDynamicButton>().label = "<";
            btn.GetComponent<UIDynamicButton>().button.onClick.AddListener(() =>
            {
                popup.popup.SetPreviousValue();
            });
            var prevBtnRect = btn.GetComponent<RectTransform>();
            prevBtnRect.pivot = new Vector2(0, 0);
            prevBtnRect.anchoredPosition = new Vector2(10f, 0);
            prevBtnRect.sizeDelta = new Vector2(0f, 0f);
            prevBtnRect.offsetMin = new Vector2(5f, 5f);
            prevBtnRect.offsetMax = new Vector2(80f, 70f);
            prevBtnRect.anchorMin = new Vector2(0f, 0f);
            prevBtnRect.anchorMax = new Vector2(0f, 0f);
        }

        {
            var btn = Instantiate(manager.configurableButtonPrefab);
            btn.SetParent(popup.transform, false);
            Destroy(btn.GetComponent<LayoutElement>());
            btn.GetComponent<UIDynamicButton>().label = ">";
            btn.GetComponent<UIDynamicButton>().button.onClick.AddListener(() =>
            {
                popup.popup.SetNextValue();
            });
            var prevBtnRect = btn.GetComponent<RectTransform>();
            prevBtnRect.pivot = new Vector2(0, 0);
            prevBtnRect.anchoredPosition = new Vector2(10f, 0);
            prevBtnRect.sizeDelta = new Vector2(0f, 0f);
            prevBtnRect.offsetMin = new Vector2(82f, 5f);
            prevBtnRect.offsetMax = new Vector2(157f, 70f);
            prevBtnRect.anchorMin = new Vector2(0f, 0f);
            prevBtnRect.anchorMax = new Vector2(0f, 0f);
        }

        return popup;

    }

    public void showGUI1()
    {

        // CreateButton("Resync gpu").button.onClick.AddListener( () => {
        //   modified.ForEach(m=>m.UpdateData());
        // });

        // SuperController.LogMessage($"in show gui 1");


        toggleshow = CreateToggle(boolshow);

        buttoncustomize = CreateButton("Customize",true);
        buttoncustomize.button.onClick.AddListener(()=> 
        { 
            destroyGUI1();
            showGUI2();
        });


        toggleseethrough = CreateToggle(boolseethrough);

        slideralpha = CreateSlider(alpha);

        buttonFit = CreateButton("Guess fit", true);
        buttonFit.button.onClick.AddListener(()=> { guessFit(); });
        fitconfirm = false;

        buttonLR = CreateButton("Copy L to R",true);
        buttonLR.button.onClick.AddListener( () => {mirrorData(false);} );
        LRconfirm = false;
        buttonRL = CreateButton("Copy R to L",true);
        buttonRL.button.onClick.AddListener( () => {mirrorData(true);} );
        RLconfirm = false;

        JSONStorableString helplabel = new JSONStorableString("HLAP", 
                    "<b>Supplementary Colliders Beta :</b> Adds several hidden capsule objects to a Person, filling in areas with missing cloth and hair collisions. \n\n"
                +   "<b>Usage :</b> Use the <b>Guess Fit</b> button for an estimated fit, then adjust as needed all the colliders on one side (names start with <b>l</b>* or <b>r</b>*)"
                +   " as well as the middle colliders (<b>m</b>*), then copy to other side using the <b>L to R</b> or <b>R to L</b> buttons.\n\n"
                +   "Guessing a fit tries to position the capsules based on only a few vertices and is not exact, in particular with extreme morphs."
        );
        labelhelp = CreateTextField(helplabel, true);
        labelhelp.height = 457.0f;

        buttonload = CreateButton("Load Preset");
        buttonload.button.onClick.AddListener(() =>
        {
            FileManagerSecure.CreateDirectory(_savePath);
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_savePath);
            SuperController.singleton.GetMediaPathDialog(HandleLoadPreset, _saveExt, _savePath, false, true, false, null, false, shortcuts);
        });

        buttonsave = CreateButton("Save Preset");
        buttonsave.button.onClick.AddListener(() =>
        {
            FileManagerSecure.CreateDirectory(_savePath);
            var fileBrowserUI = SuperController.singleton.fileBrowserUI;
            fileBrowserUI.SetTitle("Save colliders preset");
            fileBrowserUI.fileRemovePrefix = null;
            fileBrowserUI.hideExtension = false;
            fileBrowserUI.keepOpen = false;
            fileBrowserUI.fileFormat = _saveExt;
            fileBrowserUI.defaultPath = _savePath;
            fileBrowserUI.showDirs = true;
            fileBrowserUI.shortCuts = FileManagerSecure.GetShortCutsForDirectory(_savePath);
            fileBrowserUI.browseVarFilesAsDirectories = false;
            fileBrowserUI.SetTextEntry(true);
            fileBrowserUI.Show(HandleSavePreset);
            fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + _saveExt;
            fileBrowserUI.ActivateFileNameField();
        });
    
        fakecolliderchooser = new JSONStorableStringChooser("fakecolliders", fcolNames, fcolNames[0], "Colliders" );
        fakecolliderchooser.setCallbackFunction += RefreshUI;

        popupchooser = CreateAutoPopup(fakecolliderchooser);
        RefreshUI(fakecolliderchooser.val);


    }

    public void showGUI2()
    {
        useLegs = CreateToggle(boolLegs);
        toggleConvertExistingLegs = CreateToggle(boolConvertExistingLegs,true);
        toggleAlsoCollideLegs = CreateToggle(boolAlsoCollideLegs, true);
        var spacer = CreateSpacer();
        spacer.height = useLegs.height;
        spacers.Add(spacer);

        useArms = CreateToggle(boolArms);
        toggleConvertExistingArms = CreateToggle(boolConvertExistingArms,true);
        toggleAlsoCollideArms = CreateToggle(boolAlsoCollideArms, true);
        spacer = CreateSpacer();
        spacer.height = useLegs.height;
        spacers.Add(spacer);

        useHip = CreateToggle(boolHip);
        toggleConvertExistingHip = CreateToggle(boolConvertExistingHip,true);
        toggleAlsoCollideHip = CreateToggle(boolAlsoCollideHip, true);
        spacer = CreateSpacer();
        spacer.height = useLegs.height;
        spacers.Add(spacer);        

        useTorso = CreateToggle(boolTorso);
        toggleConvertExistingTorso = CreateToggle(boolConvertExistingTorso,true);
        toggleAlsoCollideTorso = CreateToggle(boolAlsoCollideTorso, true);
        spacer = CreateSpacer();
        spacer.height = useLegs.height;
        spacers.Add(spacer);

        spacers.Add(CreateSpacer());
        toggleinterp = CreateToggle(boolinterp);
        toggleinterpNew = CreateToggle(boolinterpNew);
        toggleshowinterp = CreateToggle(boolshowinterp);        

        JSONStorableString helptxt = new JSONStorableString("HLAP2", 
                    "<b>Customization :</b> Enable or disable groups of colliders. \n\n"
                +   "<b>Interpolated :</b> Creates a series of uneditable colliders between the editable ones, to approximate a smoother surface. <b>Doubles the performance impact.</b>\n\n"
                +   "<b>New Interpolated :</b> New system of interpolated colliders, heavier but possibly smoother surface.\n\n"
                +   "<b>Caution :</b> Disabling a group will delete the colliders and their settings. They will have to be refitted when enabled again.\n\n"
                +   "Collider visibility and interpolated colliders need to be refreshed (off and on) after any changes here.\n\n"
                +   "<b>Builtin to Cloth</b> Make builtin colliders affect cloth. Edit these colliders with Collider Editor plugin. You can use less additional colliders if you have configured builtin colliders.\n\n"
                +   "<b>Collide with Other Things</b> Makes added colliders react to non-cloth collisions as well (like objects and hands). Usefull for areas where the builtin colliders can't match the actual surface.\n\n"
                +   "\n<b>I recommend using builtin colliders for the hip area instead of mine. </b>\n\n"
        );
        labelhelp2 = CreateTextField(helptxt, true);
        labelhelp2.height = 500;

        buttoncustomdone = CreateButton("Done",true);
        buttoncustomdone.button.onClick.AddListener( () =>
            {
                destroyGUI2();
                showGUI1();
            }
        );

    }

    public void destroyGUI1()
    {
        // SuperController.LogMessage($"in destroy gui 1");
        RemoveToggle(toggleshow);
        RemoveToggle(toggleseethrough);
        RemoveButton(buttoncustomize);
        RemoveSlider(slideralpha);
        RemoveButton(buttonFit);
        RemoveButton(buttonLR);
        RemoveButton(buttonRL);
        RemoveButton(buttonload);
        RemoveButton(buttonsave);
        RemoveTextField(labelhelp);
        RemovePopup(popupchooser);

        foreach( FakeCollider fcol in fakecolliders) {
            fcol.destroyGUI(this);    
        }
    }
    public void destroyGUI2()
    {
        RemoveToggle(useLegs);
        RemoveToggle(useArms);
        RemoveToggle(useHip);
        RemoveToggle(useTorso);
        RemoveToggle(toggleinterp);
        RemoveToggle(toggleinterpNew);
        RemoveToggle(toggleshowinterp);
        RemoveToggle(toggleConvertExistingLegs);
        RemoveToggle(toggleAlsoCollideLegs);
        RemoveToggle(toggleConvertExistingArms);
        RemoveToggle(toggleAlsoCollideArms);
        RemoveToggle(toggleConvertExistingHip);
        RemoveToggle(toggleAlsoCollideHip);
        RemoveToggle(toggleConvertExistingTorso);
        RemoveToggle(toggleAlsoCollideTorso);
        RemoveTextField(labelhelp2);
        RemoveButton(buttoncustomdone);
        spacers.ForEach(RemoveSpacer);
        spacers.Clear();
    }

    public void guessFit(bool skipcheck=false)
    {
        if (fitconfirm || skipcheck) {
            Atom p = GetContainingAtom();
            DAZSkinV2 skin = p.GetComponentInChildren<DAZCharacter>().skin;
            Vector3[] autoverts = new Vector3[25000];
            skin.smoothedVertsBuffer.GetData(autoverts);

            foreach (FakeCollider fcol in fakecolliders) 
            {
                fcol.Autofit(autoverts);
            }
            mirrorDataInner(false);
        }
        if (!skipcheck)
            fitconfirm = buttonState(buttonFit, !fitconfirm);

    }

    public void interpFit() {
      if (boolinterp.val) interpFitOld();
      if (boolinterpNew.val) interpFitNew();
    }

    public void interpFitOld() {
        // SuperController.LogMessage($"in interpfit {interpolatedcolliders.Count()}");
        int i=0;
        foreach(FakeCollider fcol in interpolatedcolliders)
        {
            fcol.InterpolatedFit(fakecolliders);
            i++;
        }
        // SuperController.LogMessage($"end interpfit {i}");
    }

    public void interpFitNew() {
        // SuperController.LogMessage($"in interpfit new {interpolatedcollidersNew.Count()}");
        int i=0;
        foreach(FakeCollider fcol in interpolatedcollidersNew)
        {
            fcol.InterpolatedFitNew(fakecolliders);
            i++;
        }
        // SuperController.LogMessage($"end interpfit new {i}");
    }



    public void FixedUpdate() {
    }
    public void Update() {
    }

    public void RefreshUI(string fc) {
        LRconfirm = buttonState(buttonLR, false);
        RLconfirm = buttonState(buttonRL, false);
        fitconfirm = buttonState(buttonFit, false);
        foreach( FakeCollider fcol in fakecolliders) {
            fcol.destroyGUI(this);    
            if (fc == fcol.name || fc=="all") {
                fcol.createGUI(this);
            }
        }

    }

    public bool buttonState(UIDynamicButton but, bool confirm) {
        if (confirm)
        {
            // convert button to confirm button, return true flag 
            but.buttonColor = Color.green;
            string text = "Confirm " + but.buttonText.text;
            but.buttonText.text = text;
            return true;
        }
        else
        {
            // convert button back to regular button, return false flag 
            but.buttonColor = Color.white;
            string text = but.buttonText.text.Replace("Confirm ","");
            but.buttonText.text = text;
            return false;
        }

    }


    public void mirrorData(bool RtoL = false) {

        // reset both buttons
    
        // set for confirm the correct button
        if (!RtoL)
        {
            RLconfirm = buttonState(buttonRL, false);
            if (LRconfirm) {
                mirrorDataInner(RtoL);
            }
            LRconfirm = buttonState(buttonLR, !LRconfirm); 
        }
        else 
        {
            LRconfirm = buttonState(buttonLR, false);
            if (RLconfirm) {
                mirrorDataInner(RtoL);
            }
            RLconfirm = buttonState(buttonRL, !RLconfirm); 
        }
    }


    public void mirrorDataInner(bool RtoL = false) 
    {
        
        
        string sc = RtoL ? "r" : "l";
        string dc = RtoL ? "l" : "r";
    
        foreach(FakeCollider s_fcol in fakecolliders) 
        {
            if (s_fcol.name.StartsWith(sc)) 
            {
                string s_name = s_fcol.name;
                foreach(FakeCollider d_fcol in fakecolliders) 
                {

                    if (d_fcol.name.StartsWith(dc) && d_fcol.name.Substring(1)== s_fcol.name.Substring(1)) 
                    {
                        d_fcol.mirrorData(s_fcol);
                    }
                }
                
            }
        }
    }


    public JSONClass CollectJSON(JSONClass jc)
    {
        foreach( FakeCollider fcol in fakecolliders) 
        {
            jc[fcol.name] = fcol.GetJSON();

        } 
        return jc;
    }

    public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
    {
        JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
        if (includePhysical || forceStore)
        {
            needsStore = true;
            jc["version"].AsInt = 3;
            jc["uselegs"].AsBool = boolLegs.val;
            jc["usearms"].AsBool = boolArms.val;
            jc["usehip"].AsBool = boolHip.val;
            jc["usetorso"].AsBool = boolTorso.val;
            jc["useinterp"].AsBool = boolinterp.val;
            jc = CollectJSON(jc);

        }
        return jc;
    }

    private void HandleLoadPreset(string path)
    {
        if (string.IsNullOrEmpty(path))
            return;
        destroyGUI1();
        removeFcolAtoms();
        removeInterpolated();
        JSONClass jc = (JSONClass)LoadJSON(path);
        manualJC = jc;
        int version = jc["version"].AsInt;
        SuperController.LogMessage($"Version is {version}");        
        if (version==3) {
            boolLegs.valNoCallback = jc["uselegs"].AsBool;
            boolArms.valNoCallback = jc["usearms"].AsBool;
            boolHip.valNoCallback = jc["usehip"].AsBool;
            boolTorso.valNoCallback = jc["usetorso"].AsBool;
            boolinterp.valNoCallback = jc["useinterp"].AsBool;    

            boolConvertExistingLegs.val = jc["builtinlegs"].AsBool;
            boolConvertExistingArms.val = jc["builtinarms"].AsBool;
            boolConvertExistingHip.val = jc["builtinhip"].AsBool;
            boolConvertExistingTorso.val = jc["builtintorso"].AsBool;
            boolAlsoCollideLegs.val = jc["alsocollidelegs"].AsBool;
            boolAlsoCollideArms.val = jc["alsocollidearms"].AsBool;
            boolAlsoCollideHip.val = jc["alsocollidehip"].AsBool;
            boolAlsoCollideTorso.val = jc["alsocollidetorso"].AsBool;


        } else if (version==2) 
        {
            boolLegs.valNoCallback = jc["uselegs"].AsBool;
            boolArms.valNoCallback = jc["usearms"].AsBool;
            boolHip.valNoCallback = jc["usehip"].AsBool;
            boolTorso.valNoCallback = jc["usetorso"].AsBool;
            boolinterp.valNoCallback = jc["useinterp"].AsBool;    
        }
        else 
        {
            boolLegs.valNoCallback = true;
            boolArms.valNoCallback = true;
            boolHip.valNoCallback = true;
            boolTorso.valNoCallback = true;
            boolinterp.valNoCallback = false;    
        }
        CreateAll(version<3);

    }
    private void HandleSavePreset(string path)
    {
        SuperController.singleton.fileBrowserUI.fileFormat = null;
        if (string.IsNullOrEmpty(path)) return;
        if (!path.ToLower().EndsWith($".{_saveExt}")) path += $".{_saveExt}";

        var jc = new JSONClass();
        jc["version"].AsInt = 3;
        jc["uselegs"].AsBool = boolLegs.val;
        jc["usearms"].AsBool = boolArms.val;
        jc["usehip"].AsBool = boolHip.val;
        jc["usetorso"].AsBool = boolTorso.val;
        jc["useinterp"].AsBool = boolinterp.val;
        jc["builtinlegs"].AsBool = boolConvertExistingLegs.val;
        jc["builtinarms"].AsBool = boolConvertExistingArms.val;
        jc["builtinhip"].AsBool = boolConvertExistingHip.val;
        jc["builtintorso"].AsBool = boolConvertExistingTorso.val;
        jc["alsocollidelegs"].AsBool = boolAlsoCollideLegs.val;
        jc["alsocollidearms"].AsBool = boolAlsoCollideArms.val;
        jc["alsocollidehip"].AsBool = boolAlsoCollideHip.val;
        jc["alsocollidetorso"].AsBool = boolAlsoCollideTorso.val;

        CollectJSON(jc);
        SaveJSON(jc, path);
    }

    public void manualRestore(JSONClass jc, bool isoldversion=false) {
        if (jc == null)  {
          jc = manualJC;
          manualJC = null;
        }
        if (jc == null)  {
          guessFit(true); // if there's nothing to load, do an autofit instead
          return;
        }

        foreach( FakeCollider fcol in fakecolliders) {
            // SuperController.LogMessage($"restoring {fcol.name}");
          if (jc.HasKey(fcol.name)) {
              fcol.RestoreFromJSON(jc[fcol.name].AsObject, correctoldversion:isoldversion);
          }
        }
    }

    public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
        // SuperController.LogMessage($"INSIDE RESTORE {restorePhysical}, {restoreAppearance}, {setMissingToDefault}");
        destroyGUI1();
        base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
        int version = 3;
        if (!base.physicalLocked && restorePhysical)
        {

            version = jc["version"].AsInt;
            SuperController.LogMessage($"Version is {version}");
            if (version == 3) {
  
            } else if (version==2) {
                boolLegs.val = jc["uselegs"].AsBool;
                boolArms.val = jc["usearms"].AsBool;
                boolHip.val = jc["usehip"].AsBool;
                boolTorso.val = jc["usetorso"].AsBool;
                boolinterp.val = jc["useinterp"].AsBool;    
            } else {
                boolLegs.val = true;
                boolArms.val = true;
                boolHip.val = true;
                boolTorso.val = true;
                boolinterp.val = false;
            }
            manualJC = jc;
        }
        manualRestore(jc, version<3);
        showGUI1();
    }

    public void OnEnable() {
        try
        {
            SuperController.LogMessage($"{nameof(UYCol)} enabled");
        }
        catch (Exception e)
        {
            SuperController.LogError($"{nameof(UYCol)}.{nameof(OnEnable)}: {e}");
        }
    }

    public void OnDisable() {
        try
        {
            SuperController.LogMessage($"{nameof(UYCol)} disabled");
        }
        catch (Exception e)
        {
            SuperController.LogError($"{nameof(UYCol)}.{nameof(OnDisable)}: {e}");
        }
    }

    public void OnDestroy() {
        // foreach( FakeCollider fcol in fakecolliders) 
        // {
        //     fcol.Destroy();            
        // }
        removeFcolAtoms();
        removeInterpolated();
        removeInterpolatedNew();
        RestoreModifiedExisting(legRigidbodies);
        RestoreModifiedExisting(armRigidbodies);
        RestoreModifiedExisting(hipRigidbodies);
        RestoreModifiedExisting(torsoRigidbodies);

        try
        {
            SuperController.LogMessage($"{nameof(UYCol)} destroyed");
        }
        catch (Exception e)
        {
            SuperController.LogError($"{nameof(UYCol)}.{nameof(OnDestroy)}: {e}");
        }
    }
}
